#
#   MLCV Winter 2016
#   Homework Assignment 06 - Sampling methods
#
#   Exercise 1) Particle Filter for MCL
#
#   (c) John Chiotellis
###########################################################################
from time import time
import numpy as np
from math import cos, sin
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation


class Localization:
    def __init__(self, worldMap, robot, particleFilter):
        self.map = worldMap
        self.robot = robot
        self.pf = particleFilter
        self.fig, self.ax = plt.subplots()

        # Set up plot parameters
        self.ax.grid(True)
        self.lmpos = self.ax.scatter([], [], c='k', marker='s', s=100)
        self.ppos = self.ax.scatter([], [], c='none', edgecolors='b', s=30, marker='o')
        self.pori, = self.ax.plot([], [], 'k-')
        self.rpos = self.ax.scatter([], [], c='r', edgecolors='none', s=80, marker='o')
        self.rori, = self.ax.plot([], [], 'w-')
        self.anim = None

    def loop(self):
        self.anim = FuncAnimation(fig=self.fig, func=self.update, interval=100, blit=True,
                                  init_func=self.init_plot)
        plt.show()

    def init_plot(self):
        print('Initial Drawing...', end='')
        t1 = time()

        # Draw landmarks
        self.lmpos.set_offsets([[lm.x, lm.y] for lm in self.map.landmarks])
        self.ax.axis([0, self.map.size[0], 0, self.map.size[1]])
        print('Took %.4f s' % (time() - t1))
        return self.lmpos, self.rpos, self.rori, self.ppos, self.pori,

    def update(self, dummy):
        print('Drawing...', end='')
        t1 = time()
        robot = self.robot
        particles = self.pf.particles

        turn = 0.1
        motion = 5
        # Move robot
        robot.move(turn, motion)
        # Move particles
        self.pf.action_update(turn, motion)

        # Sense with robot
        z = robot.sense()
        # Evaluate measurement likelihood
        self.pf.measurement_update(z)

        # Resample
        self.pf.resample()

        # Draw particles positions
        self.ppos.set_offsets([[p.x, p.y] for p in particles])

        # Draw particles orientations (interleave nans to cut line segments)
        self.pori.set_data([[[p.x, p.x + cos(p.theta)], [np.nan, np.nan]] for p in particles],
                           [[[p.y, p.y + sin(p.theta)], [np.nan, np.nan]] for p in particles])

        # Draw robot position
        self.rpos.set_offsets([robot.x, robot.y])

        # Draw robot orientation
        self.rori.set_data([robot.x, robot.x + cos(robot.theta)],
                           [robot.y, robot.y + sin(robot.theta)])

        self.pf.eval(robot)

        print('Took %.4f s' % (time() - t1))
        return self.lmpos, self.rpos, self.rori, self.ppos, self.pori,
