#
#   MLCV Winter 2016
#   Homework Assignment 06 - Sampling methods
#
#   Exercise 1) Particle Filter for MCL
#
#   (c) John Chiotellis
###########################################################################
import numpy as np
from time import time
from Robot import Robot
from copy import copy


class ParticleFilter:
    def __init__(self, N, world, sigma_move, sigma_turn, sigma_sense):
        self.N = N
        self.world = world
        self.particles = []

        for i in range(N):
            r = Robot(world=world)
            r.set_noise(sigma_move, sigma_turn, sigma_sense)
            self.particles.append(r)

        self.weights = np.ones(self.N) / self.N

    def reset_weights(self):
        self.weights = np.ones(self.N) / self.N

    def action_update(self, turn, motion):
        print('Moving %d particles...' % self.N, end='')
        t1 = time()
        [p.move(turn, motion) for p in self.particles]
        print('Took %.4f s' % (time()-t1))

    def measurement_update(self, measurement):
        print('Computing particles weights...', end='')
        t1 = time()
        self.weights = np.array([p.measurement_prob(measurement) for p in self.particles], copy=False)

        if max(self.weights) == 0:
            self.reset_weights()

        print('Took %.4f s' % (time() - t1 ))

    def resample(self):
        '''''
        Low variance resampling
        '''''
        print('Resampling...')
        t1 = time()
        new_particles = [None]*self.N
        W = sum(self.weights)
        u = np.random.rand()*W / self.N
        i, cur_weight_sum = 0, self.weights[0]
        stepsize = 1*W / self.N
        for j in range(self.N):
            # print('Resampling particle %d/%d' % (j, self.N)) if j % 100 == 0 else None
            while cur_weight_sum < u:
                i = (i+1) % self.N
                cur_weight_sum += self.weights[i]
            new_particles[j] = copy(self.particles[i])
            u += stepsize

        self.particles = new_particles
        self.reset_weights()
        print('Took %.4f s' % (time() - t1))

    def eval(self, groundtruth):
        print('Computing mean error...', end='')
        t1 = time()
        err_sum = 0
        wx, wy = self.world.size
        for p in self.particles:
            adx = abs(p.x - groundtruth.x)
            ady = abs(p.y - groundtruth.y)
            dx = min(adx, wx - adx)
            dy = min(ady, wy - ady)
            err_sum += np.sqrt(dx ** 2 + dy ** 2)

        mean_error = err_sum / self.N
        print('Took %.4f s\nMean error= %.4f\n' % (time() - t1, mean_error))
        return mean_error
