#
#   MLCV Winter 2016
#   Homework Assignment 06 - Sampling methods
#
#   Exercise 1) Particle Filter for MCL
#
#   (c) John Chiotellis
###########################################################################
from scipy import stats
import numpy as np
from numpy import pi, cos, sin


class Point:
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y


class Robot:
    def __init__(self, world, x=None, y=None, theta=None, sigma_move=0, sigma_turn=0, sigma_sense=0):
        self.x = x if x is not None else np.random.rand() * world.size[0]
        self.y = y if y is not None else np.random.rand() * world.size[1]
        self.theta = theta if theta is not None else np.random.rand() * 2 * pi
        self.sigma_move = sigma_move
        self.sigma_turn = sigma_turn
        self.sigma_sense = sigma_sense
        self.world = world

    def set_pose(self, x=0, y=0, theta=0):
        self.x = x
        self.y = y
        self.theta = theta

    def set_noise(self, sigma_move=0, sigma_turn=0, sigma_sense=0):
        self.sigma_move = sigma_move
        self.sigma_turn = sigma_turn
        self.sigma_sense = sigma_sense

    def sense(self):
        measurement = []
        distances = self.world.distances_to_landmarks(self)
        for dist in distances:
            measurement.append(dist + np.random.normal(loc=0, scale=self.sigma_sense))
        return np.array(measurement, copy=False)

    def move(self, turn, move):
        if move < 0:
            raise ValueError('Robot cannot move backwards.')

        # Turn and add turn noise
        theta_new = self.theta + turn + np.random.normal(loc=0, scale=self.sigma_turn)
        self.theta = theta_new % (2*pi)

        # Move and add move noise
        dist = move + np.random.normal(loc=0, scale=self.sigma_move)
        x_new = self.x + (dist*cos(self.theta))
        y_new = self.y + (dist*sin(self.theta))

        self.x = x_new % self.world.size[0]
        self.y = y_new % self.world.size[1]

    def measurement_prob(self, measurement):
        prob = 1.0
        distances = self.world.distances_to_landmarks(self)
        for meas, dist in zip(measurement, distances):
            prob *= stats.norm.pdf(meas, loc=dist, scale=self.sigma_sense)
        return prob
