#
#   MLCV Winter 2016
#   Homework Assignment 06 - Sampling methods
#
#   Exercise 1) Particle Filter for MCL
#
#   (c) John Chiotellis
###########################################################################
from Robot import Robot, Point
from WorldMap import WorldMap
from ParticleFilter import ParticleFilter
from Localization import Localization
from math import pi
import numpy as np


def main():
    world_size = (100, 100)
    landmark_locations = [(20, 20), (20, 80), (80, 20), (80, 80)]
    landmarks = [Point(x, y) for x, y in landmark_locations]
    world = WorldMap(world_size, landmarks)

    # Initialize noise values
    sigma_move  = 0.1
    sigma_turn  = 0.05
    sigma_sense = 3

    # For reproducibility
    np.random.seed(seed=1)

    # Initialize robot
    robot = Robot(x=30, y=30, theta=pi/6, world=world)
    robot.set_noise(sigma_move, sigma_turn, sigma_sense)

    # Initialize particles
    N = 1000
    pf = ParticleFilter(N, world, sigma_move, sigma_turn, sigma_sense)

    # Compute initial error
    pf.eval(robot)

    # Start localization - animation loop
    mcl = Localization(world, robot, pf)
    mcl.loop()

if __name__ == '__main__':
    main()
